@@ -1,5 +1,5 @@
 /* This file is part of GNU Pies
-   Copyright (C) 2009 Sergey Poznyakoff
+   Copyright (C) 2009-2016 Sergey Poznyakoff
   
    GNU Pies is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -25,7 +25,6 @@
 #include <netinet/in.h>
 #include <arpa/inet.h>
 #include <netdb.h>
-#include <hash.h>
 
 struct pies_sockaddr
 {
@@ -34,51 +33,134 @@ struct pies_sockaddr
   struct sockaddr sa;
 };
 
+enum name_match
+  {
+    match_none,
+    match_user_name,
+    match_group_name
+  };
+
 struct acl_entry
 {
   grecs_locus_t locus;
   int allow;
   int authenticated;
   pies_acl_t acl;
-  gl_list_t groups;
-  gl_list_t sockaddrs;
+  enum name_match name_match;
+  char **names;
+  struct grecs_list *sockaddrs;
 };
 
 struct pies_acl
 {
   char *name;
+  size_t refcnt;
   grecs_locus_t locus;
-  gl_list_t list;
+  struct grecs_list *list;
 };
    
 
 
 /* ACL creation */
+void
+grecs_locus_point_copy (struct grecs_locus_point *dst,
+			struct grecs_locus_point *src)
+{
+  dst->file = grecs_strdup (src->file);
+  dst->line = src->line;
+  dst->col = src->col;
+}
+
+void
+grecs_locus_copy (struct grecs_locus *dst, struct grecs_locus *src)
+{
+  grecs_locus_point_copy (&dst->beg, &src->beg);
+  grecs_locus_point_copy (&dst->end, &src->end);
+}
+
+void
+grecs_locus_point_free (struct grecs_locus_point *p)
+{
+  grecs_free (p->file);
+}
+
+void
+grecs_locus_free (struct grecs_locus *loc)
+{
+  grecs_locus_point_free (&loc->beg);
+  grecs_locus_point_free (&loc->end);
+}
+
+static void
+acl_free_entry (void *p)
+{
+  struct acl_entry *ent = p;
+  pies_acl_free (ent->acl);
+  grecs_locus_free (&ent->locus);
+  grecs_list_free (ent->sockaddrs);
+  if (ent->names)
+    {
+      size_t i;
+
+      for (i = 0; ent->names[i]; i++)
+	free (ent->names[i]);
+      free (ent->names);
+    }
+  free (ent);
+}
+
+void
+pies_acl_use (pies_acl_t acl)
+{
+  ++acl->refcnt;
+}
 
 pies_acl_t
 pies_acl_create (const char *name, grecs_locus_t *locus)
 {
-  pies_acl_t acl = xmalloc (sizeof (acl[0]));
-  acl->name = name ? xstrdup (name) : NULL;
-  acl->locus = *locus;
-  acl->list = gl_list_create_empty(&gl_linked_list_implementation,
-				   NULL,
-				   NULL,
-				   NULL,
-				   false);
+  pies_acl_t acl = grecs_malloc (sizeof (acl[0]));
+  acl->name = name ? grecs_strdup (name) : NULL;
+  acl->refcnt = 0;
+  grecs_locus_copy (&acl->locus, locus);
+  acl->list = grecs_list_create ();
+  acl->list->free_entry = acl_free_entry;
+  pies_acl_use (acl);
   return acl;
 }
 
+void
+pies_acl_destroy (pies_acl_t *pacl)
+{
+  if (pacl && *pacl && (*pacl)->refcnt)
+    {
+      pies_acl_t acl = *pacl;
+      if (--acl->refcnt == 0)
+	{
+	  free (acl->name);
+	  grecs_locus_free (&acl->locus);
+	  grecs_list_free (acl->list);
+	  free (acl);
+	  *pacl = NULL;
+	}
+    }
+}
+
+void
+pies_acl_free (pies_acl_t acl)
+{
+  pies_acl_destroy (&acl);
+}
+
 static struct pies_sockaddr *
 create_acl_sockaddr (int family, int len)
 {
-  struct pies_sockaddr *p = xzalloc (sizeof (*p));
+  struct pies_sockaddr *p = grecs_zalloc (sizeof (*p));
   p->salen = len;
   p->sa.sa_family = family;
   return p;
 }
 
-/* allow|deny [all|authenticated|group <grp: list>]
+/* allow|deny [all|authenticated|group <grp: list>|user <usr: list>]
               [acl <name: string>] [from <addr: list>] */
 
 static int
@@ -86,7 +168,7 @@ _parse_token (struct acl_entry *entry, const grecs_value_t *value)
 {
   if (strcmp (value->v.string, "all") == 0
       || strcmp (value->v.string, "any") == 0)
-    /* FIXME: Nothing? */ ;
+    /* nothing */ ;
   else if (strcmp (value->v.string, "auth") == 0
 	   || strcmp (value->v.string, "authenticated") == 0)
     entry->authenticated = 1;
@@ -115,7 +197,7 @@ _parse_sockaddr (struct acl_entry *entry, const grecs_value_t *value)
       if (len >= sizeof (s_un->sun_path))
 	{
 	  grecs_error (&entry->locus, 0,
-		       _("socket name too long: `%s'"), string);
+		       _("socket name too long: %s"), string);
 	  return 1;
 	}
       sptr = create_acl_sockaddr (AF_UNIX, sizeof (s_un));
@@ -138,7 +220,7 @@ _parse_sockaddr (struct acl_entry *entry, const grecs_value_t *value)
 	  if (!hp)
 	    {
 	      grecs_error (&entry->locus, 0,
-			   _("cannot resolve host name: `%s'"), string);
+			   _("cannot resolve host name: %s"), string);
 	      if (p)
 		*p = '/';
 	      return 1;
@@ -175,38 +257,57 @@ _parse_sockaddr (struct acl_entry *entry, const grecs_value_t *value)
 	      if (inet_aton (p, &addr) == 0)
 		{
 		  grecs_error (&entry->locus, 0,
-			       _("invalid netmask: `%s'"), p);
+			       _("invalid netmask: %s"), p);
 		  return 1;
 		}
 	      sptr->netmask = addr.s_addr;
 	    }
 	  else
 	    {
-	      grecs_error (&entry->locus, 0, _("invalid netmask: `%s'"), p);
+	      grecs_error (&entry->locus, 0, _("invalid netmask: %s"), p);
 	      return 1;
 	    }
 	}
       else
 	sptr->netmask = 0xfffffffful;
     }
-  gl_list_add_last (entry->sockaddrs, sptr);
+  grecs_list_append (entry->sockaddrs, sptr);
   return 0;
 }
 
 static int
-_parse_from (struct acl_entry *entry, size_t argc, const grecs_value_t *argv)
+sockaddr_cmp (void const *a, void const *b)
+{
+  struct pies_sockaddr const *ap = a;
+  struct pies_sockaddr const *bp = b;
+
+  if (ap->netmask != bp->netmask)
+    return 1;
+  if (ap->salen != bp->salen)
+    return 1;
+  return memcmp (&ap->sa, &bp->sa, ap->salen);
+}
+  
+static void
+sockaddr_free (void *p)
+{
+  free (p);
+}
+
+static int
+_parse_from (struct acl_entry *entry, size_t argc, grecs_value_t **argv)
 {
   if (argc == 0)
     return 0;
-  else if (argv->type == GRECS_TYPE_LIST)
+  else if (argv[0]->type == GRECS_TYPE_LIST)
     {
-      grecs_error (&entry->locus, 0, _("expected `from', but found list"));
+      grecs_error (&entry->locus, 0, _("expected \"from\", but found list"));
       return 1;
     }
-  else if (strcmp (argv->v.string, "from"))
+  else if (strcmp (argv[0]->v.string, "from"))
     {
-      grecs_error (&entry->locus, 0, _("expected `from', but found `%s'"),
-		    argv->v.string);
+      grecs_error (&entry->locus, 0, _("expected \"from\", but found \"%s\""),
+		    argv[0]->v.string);
       return 1;
     }
   argc--;
@@ -215,46 +316,43 @@ _parse_from (struct acl_entry *entry, size_t argc, const grecs_value_t *argv)
   if (argc == 0)
     {
       grecs_error (&entry->locus, 0,
-		   _("unexpected end of statement after `from'"));
+		   _("unexpected end of statement after \"from\""));
       return 1;
     }
 
-  entry->sockaddrs = gl_list_create_empty(&gl_linked_list_implementation,
-					  NULL,
-					  NULL,
-					  NULL,
-					  false);
-  if (argv->type == GRECS_TYPE_STRING)
+  entry->sockaddrs = grecs_list_create ();
+  entry->sockaddrs->cmp = sockaddr_cmp;
+  entry->sockaddrs->free_entry = sockaddr_free;
+  if (argv[0]->type == GRECS_TYPE_STRING)
     {
-      if (_parse_sockaddr (entry, argv))
+      if (_parse_sockaddr (entry, argv[0]))
 	return 1;
     }
   else
     {
-      gl_list_iterator_t itr = gl_list_iterator (argv->v.list);
-      const void *p;
       int rc = 0;
-      while (gl_list_iterator_next (&itr, &p, NULL))
-	rc += _parse_sockaddr (entry, (const grecs_value_t*) p);
-      gl_list_iterator_free (&itr);
+      struct grecs_list_entry *ep;
+
+      for (ep = argv[0]->v.list->head; ep; ep = ep->next)
+	rc += _parse_sockaddr (entry, (const grecs_value_t*) ep->data);
       if (rc)
 	return rc;
     }
 
   if (argc - 1)
     {
-      grecs_warning (&entry->locus, 0, _("junk after `from' list"));
+      grecs_warning (&entry->locus, 0, _("junk after from-list"));
       return 1;
     }
   return 0;
 }
 
 static int
-_parse_sub_acl (struct acl_entry *entry, size_t argc, grecs_value_t *argv)
+_parse_sub_acl (struct acl_entry *entry, size_t argc, grecs_value_t **argv)
 {
   if (argc == 0)
     return 0;
-  if (strcmp (argv->v.string, "acl") == 0)
+  if (strcmp (argv[0]->v.string, "acl") == 0)
     {
       argc--;
       argv++;
@@ -265,21 +363,23 @@ _parse_sub_acl (struct acl_entry *entry, size_t argc, grecs_value_t *argv)
 	  return 1;
 	}
 
-      if (argv->type != GRECS_TYPE_STRING)
+      if (argv[0]->type != GRECS_TYPE_STRING)
 	{
 	  grecs_error (&entry->locus, 0,
 		       _("expected string, but found list"));
 	  return 1;
 	}
 
-      entry->acl = pies_acl_lookup (argv->v.string);
+      entry->acl = pies_acl_lookup (argv[0]->v.string);
 
       if (!entry->acl)
 	{
-	  grecs_error (&entry->locus, 0, _("ACL not defined: `%s'"),
-			argv->v.string);
+	  grecs_error (&entry->locus, 0, _("ACL not defined: %s"),
+		       argv[0]->v.string);
 	  return 1;
 	}
+      pies_acl_use (entry->acl);
+      
       argc--;
       argv++;
     }
@@ -287,29 +387,41 @@ _parse_sub_acl (struct acl_entry *entry, size_t argc, grecs_value_t *argv)
 }
 
 static int
-_parse_group (struct acl_entry *entry, size_t argc, grecs_value_t * argv)
+_parse_group (struct acl_entry *entry, size_t argc, grecs_value_t **argv)
 {
-  if (strcmp (argv->v.string, "group") == 0)
+  if (strcmp (argv[0]->v.string, "group") == 0)
+    entry->name_match = match_group_name;
+  else if (strcmp (argv[0]->v.string, "user") == 0)
+    entry->name_match = match_user_name;
+  else
+    entry->name_match = match_none;
+
+  if (entry->name_match != match_none)
     {
       argc--;
       argv++;
       if (argc == 0)
 	{
 	  grecs_error (&entry->locus, 0,
-		       _("expected group list, but found end of statement"));
+		       _("expected identity list, but found end of statement"));
 	  return 1;
 	}
-      if (argv->type == GRECS_TYPE_STRING)
+      if (argv[0]->type == GRECS_TYPE_STRING)
 	{
-	  entry->groups = gl_list_create_empty(&gl_linked_list_implementation,
-					       NULL,
-					       NULL,
-					       NULL,
-					       false);
-	  gl_list_add_last (entry->groups, (void *) argv->v.string);
+	  entry->names = grecs_calloc (2, sizeof (entry->names[0]));
+	  entry->names[0] = grecs_strdup (argv[0]->v.string);
+	  entry->names[1] = NULL;
 	}
       else
-	entry->groups = argv->v.list;
+	{
+	  size_t i;
+	  struct grecs_list_entry *ep;
+	  entry->names = grecs_calloc (argv[0]->v.list->count + 1,
+				       sizeof (entry->names[0]));
+	  for (i = 0, ep = argv[0]->v.list->head; ep; ep = ep->next, ++i)
+	    entry->names[i] = grecs_strdup (ep->data);
+	  entry->names[i] = NULL;
+	}
       argc--;
       argv++;
     }
@@ -317,11 +429,11 @@ _parse_group (struct acl_entry *entry, size_t argc, grecs_value_t * argv)
 }
 
 static int
-_parse_acl (struct acl_entry *entry, size_t argc, grecs_value_t *argv)
+_parse_acl (struct acl_entry *entry, size_t argc, grecs_value_t **argv)
 {
-  if (assert_grecs_value_type (&entry->locus, argv, GRECS_TYPE_STRING))
+  if (assert_grecs_value_type (&entry->locus, argv[0], GRECS_TYPE_STRING))
     return 1;
-  else if (_parse_token (entry, argv) == 0)
+  else if (_parse_token (entry, argv[0]) == 0)
     return _parse_sub_acl (entry, argc - 1, argv + 1);
   else
     return _parse_group (entry, argc, argv);
@@ -331,35 +443,66 @@ int
 parse_acl_line (grecs_locus_t *locus, int allow, pies_acl_t acl,
 		grecs_value_t *value)
 {
-  struct acl_entry *entry = xzalloc (sizeof (*entry));
+  struct acl_entry *entry = grecs_zalloc (sizeof (*entry));
 
-  entry->locus = *locus;
+  grecs_locus_copy (&entry->locus, locus);
   entry->allow = allow;
 
-  switch (value->type)
-    {
-    case GRECS_TYPE_STRING:
-      if (_parse_token (entry, value))
-	{
-	  grecs_error (&entry->locus, 0, _("unknown word `%s'"),
-		       value->v.string);
+  if (value)
+    switch (value->type)
+      {
+      case GRECS_TYPE_STRING:
+	if (_parse_token (entry, value))
+	  {
+	    grecs_error (&entry->locus, 0, _("unknown word: %s"),
+			 value->v.string);
+	    return 1;
+	  }
+	break;
+	
+      case GRECS_TYPE_ARRAY:
+	if (_parse_acl (entry, value->v.arg.c, value->v.arg.v))
 	  return 1;
-	}
-      break;
-
-    case GRECS_TYPE_ARRAY:
-      if (_parse_acl (entry, value->v.arg.c, value->v.arg.v))
+	break;
+	
+      case GRECS_TYPE_LIST:
+	grecs_error (locus, 0, _("unexpected list"));
 	return 1;
-      break;
-
-    case GRECS_TYPE_LIST:
-      grecs_error (locus, 0, _("unexpected list"));
-      return 1;
-    }
-  gl_list_add_last (acl->list, entry);
+      }
+  grecs_list_append (acl->list, entry);
   return 0;
 }
+
+static int
+acl_entry_cmp (void const *a, void const *b)
+{
+  struct acl_entry const *ap = a;
+  struct acl_entry const *bp = b;
+  size_t i;
+    
+  if (ap->allow != bp->allow)
+    return 1;
+  if (ap->authenticated != bp->authenticated)
+    return 1;
+  if (pies_acl_cmp (ap->acl, bp->acl))
+    return 1;
+  if (ap->name_match != bp->name_match)
+    return 1;
 
+  if (ap->names && bp->names)
+    {
+      for (i = 0; ap->names[i]; i++)
+	if (!bp->names[i] || strcmp (ap->names[i], bp->names[i]))
+	  return 1;
+      if (bp->names[i])
+	return 1;
+    }
+  else if (ap->names || bp->names)
+    return 1;
+  
+  return grecs_list_compare (ap->sockaddrs, bp->sockaddrs);
+}
+
 #define ACL_TAG_NONE     0
 #define ACL_TAG_IGNORE   1
 #define ACL_TAG_OPTIONAL 2
@@ -373,7 +516,6 @@ _acl_common_section_parser (enum grecs_callback_command cmd,
 			    int flag)
 {
   pies_acl_t acl;
-  grecs_locus_t defn_loc;
   const char *tag = NULL;
   int has_value = 0;
   
@@ -412,21 +554,37 @@ _acl_common_section_parser (enum grecs_callback_command cmd,
 	  return 1;
 	}
       acl = pies_acl_create (tag, locus);
-      if (tag && pies_acl_install (acl, &defn_loc))
-	{
-	  grecs_error (locus, 0,
-		       _("redefinition of ACL %s"),
-		       value->v.string);
-	  grecs_error (&defn_loc, 0,
-		       _("location of the previous definition"));
-	  return 1;
-	}
+      if (tag && (acl = pies_acl_install (acl)) == NULL)
+	return 1;
       if (pacl)
-	*pacl = acl;
+	{
+	  pies_acl_free (*pacl);
+	  *pacl = acl;
+	}
       break;
 
     case grecs_callback_section_end:
+      acl = *pacl;
+      if (acl->list)
+	acl->list->cmp = acl_entry_cmp;
+      break;
+      
     case grecs_callback_set_value:
+      if (assert_grecs_value_type (&value->locus, value, GRECS_TYPE_STRING))
+	return 0;
+      acl = pies_acl_lookup (value->v.string);
+      if (!acl)
+	{
+	  grecs_error (&value->locus, 0, _("ACL not defined: %s"),
+		       value->v.string);
+	  return 0;
+	}
+      pies_acl_use (acl);
+      if (pacl)
+	{
+	  pies_acl_free (*pacl);
+	  *pacl = acl;
+	}
       break;
     }
   return 0;
@@ -494,21 +652,19 @@ deny_cb (enum grecs_callback_command cmd,
 
 struct grecs_keyword acl_keywords[] = {
   /* TRANSLATORS: only words within angle brackets are translatable */
-  { "allow", N_("[all|authenticated|group <grp: list>] [from <addr: list>]"),
+  { "allow", N_("[all|authenticated|user <usr: list>|group <grp: list>] [acl <name: string>] [from <addr: list>]"),
     N_("Allow access"),
-    grecs_type_string, NULL, 0,
+    grecs_type_string, GRECS_MULT, NULL, 0,
     allow_cb },
   /* TRANSLATORS: only words within angle brackets are translatable */
-  { "deny", N_("[all|authenticated|group <grp: list>] [from <addr: list>]"),
+  { "deny", N_("[all|authenticated|user <usr: list>|group <grp: list>] [acl <name: string>] [from <addr: list>]"),
     N_("Deny access"),
-    grecs_type_string, NULL, 0,
+    grecs_type_string, GRECS_MULT, NULL, 0,
     deny_cb },
   { NULL }
 };
 
 
-
-
 /* ACL verification */
 
 #define S_UN_NAME(sa, salen) \
@@ -547,15 +703,6 @@ _check_sockaddr (struct pies_sockaddr *sptr, struct acl_input *input)
   return 0;
 }
 
-static int
-match_group (const char **groups, const char *arg)
-{
-  for (; *groups; groups++)
-    if (strcmp (*groups, arg) == 0)
-      return 1;
-  return 0;
-}
-
 static int
 _acl_check (struct acl_entry *ent, struct acl_input *input)
 {
@@ -563,18 +710,24 @@ _acl_check (struct acl_entry *ent, struct acl_input *input)
   
   if (ent->authenticated)
     {
-      result = input->user != NULL;
+      result = input->identity != NULL;
       if (!result)
 	return result;
     }
 
-  if (ent->groups)
+  switch (ent->name_match)
     {
-      const void *p;
-      gl_list_iterator_t itr = gl_list_iterator (ent->groups);
-      while (result && gl_list_iterator_next (&itr, &p, NULL))
-	result = match_group (input->groups, p);
-      gl_list_iterator_free (&itr);
+    case match_none:
+      break;
+
+    case match_user_name:
+      result = pies_identity_is_user (input->identity, ent->names);
+      if (!result)
+	return result;
+      break;
+      
+    case match_group_name:
+      result = pies_identity_is_group_member (input->identity, ent->names);
       if (!result)
 	return result;
     }
@@ -585,16 +738,15 @@ _acl_check (struct acl_entry *ent, struct acl_input *input)
 
   if (ent->sockaddrs)
     {
-      const void *p;
-      gl_list_iterator_t itr = gl_list_iterator (ent->sockaddrs);
+      struct grecs_list_entry *ep;
+
       result = 0;
-      while (gl_list_iterator_next (&itr, &p, NULL))
+      for (ep = ent->sockaddrs->head; ep; ep = ep->next)
 	{
-	  result = _check_sockaddr ((struct pies_sockaddr *)p, input);
+	  result = _check_sockaddr ((struct pies_sockaddr *)ep->data, input);
 	  if (result)
 	    break;
 	}
-      gl_list_iterator_free (&itr);
     }
 
   return result;
@@ -604,7 +756,7 @@ static int
 _acl_check_cb (struct acl_entry *ent, struct acl_input *input, int *pres)
 {
   int result = _acl_check (ent, input);
-  debug (1, ("%s:%d: %s", ent->locus.file, ent->locus.line,
+  debug (1, ("%s:%d: %s", ent->locus.beg.file, ent->locus.beg.line,
 	      /* TRANSLATORS: `MATCHES' is the verb `match' in 2nd person.
 		 E.g., in French: CONCORD AVEC */
 	      result ? _("MATCHES") : _("does not match")));
@@ -622,12 +774,11 @@ pies_acl_check (pies_acl_t acl, struct acl_input *input, int result)
 {
   if (acl)
     {
-      const void *p;
-      gl_list_iterator_t itr = gl_list_iterator (acl->list);
-      while (gl_list_iterator_next (&itr, &p, NULL)
-	     && !_acl_check_cb ((struct acl_entry *)p, input, &result))
-	;
-      gl_list_iterator_free (&itr);
+      struct grecs_list_entry *ep;
+
+      for (ep = acl->list->head; ep; ep = ep->next)
+	if (_acl_check_cb ((struct acl_entry *)ep->data, input, &result))
+	  break;
     }
   return result;
 }
@@ -635,43 +786,76 @@ pies_acl_check (pies_acl_t acl, struct acl_input *input, int result)
 
 /* Hash table */
 
-static Hash_table *acl_table;
+static struct grecs_symtab *acl_table;
 
 /* Calculate the hash of a string.  */
-static size_t
-acl_hasher (void const *data, size_t n_buckets)
+static unsigned
+acl_hasher (void *data, unsigned long n_buckets)
 {
   const struct pies_acl *p = data;
-  return hash_string (p->name, n_buckets);
+  return grecs_hash_string (p->name, n_buckets);
 }
 
 /* Compare two strings for equality.  */
-static bool
+static int
 acl_compare (void const *data1, void const *data2)
 {
   const struct pies_acl *p1 = data1;
   const struct pies_acl *p2 = data2;
-  return strcasecmp (p1->name, p2->name) == 0;
+  return strcasecmp (p1->name, p2->name);
 }
 
-int
-pies_acl_install (pies_acl_t acl, grecs_locus_t * locus)
+static int
+acl_copy (void *a, void *b)
+{
+  memcpy (a, b, sizeof (struct pies_acl));
+  memset (b, 0, sizeof (struct pies_acl));
+  return 0;
+}
+
+static void
+acl_free (void *p)
+{
+  pies_acl_free (p);
+}
+
+pies_acl_t
+pies_acl_install (pies_acl_t acl)
 {
   pies_acl_t ret;
-  if (!((acl_table
-	 || (acl_table = hash_initialize (0, 0,
-					  acl_hasher,
-					  acl_compare, 0)))
-	&& (ret = hash_insert (acl_table, acl))))
-    xalloc_die ();
-
-  if (ret != acl)
+  int install = 1;
+  
+  if (!acl_table)
     {
-      if (locus)
-	*locus = ret->locus;
-      return 1;
+      acl_table = grecs_symtab_create(sizeof (struct pies_acl),
+				      acl_hasher,
+				      acl_compare,
+				      acl_copy,
+				      NULL,
+				      acl_free);
+      if (!acl_table)
+	grecs_alloc_die ();
     }
-  return 0;
+
+  ret = grecs_symtab_lookup_or_install (acl_table, acl, &install);
+
+  if (!ret)
+    {
+      logmsg (LOG_ERR, _("cannot install ACL: %s"), strerror (errno));
+      exit (1);
+    }
+  
+  if (!install)
+    {
+      grecs_error (&acl->locus, 0,
+		   _("redefinition of ACL %s"),
+		   ret->name);
+      grecs_error (&ret->locus, 0,
+		   _("location of the previous definition"));
+      ret = NULL;
+    }      
+  pies_acl_free (acl);
+  return ret;
 }
 
 pies_acl_t
@@ -681,5 +865,15 @@ pies_acl_lookup (const char *name)
   if (!acl_table)
     return NULL;
   samp.name = (char *) name;
-  return hash_lookup (acl_table, &samp);
+  return grecs_symtab_lookup_or_install (acl_table, &samp, NULL);
+}
+
+int
+pies_acl_cmp (struct pies_acl *a, struct pies_acl *b)
+{
+  if (!a)
+    return !!b;
+  else
+    return 1;
+  return grecs_list_compare (a->list, b->list);
 }
